// Copyright © 2022-2025 Gerald B. Cox
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

package mpdpost

import (
	"bufio"
	"context"
	"crypto/rand"
	"encoding/hex"
	"encoding/json"
	"fmt"
	"log/slog"
	"net"
	"net/http"
	"os"
	"strings"
	"time"
)

type Item struct {
	Track         string            `json:"Track"`
	Artist        string            `json:"Artist"`
	Album         string            `json:"Album"`
	ProviderIds   map[string]string `json:"ProviderIds"`
	PlaySessionId string            `json:"PlaySessionId"`
}

type Request struct {
	Event string `json:"Event"`
	User  string `json:"User"`
	Item  Item   `json:"Item"`
}

var (
	lastSeen    Item
	defaultUser = os.Getenv("USER")
	dispatchURL = os.Getenv("BEETBRAINZ_URL_MPD")
	mpdHost     = os.Getenv("MPD_HOST")
	mpdPort     = os.Getenv("MPD_PORT")
)

func generateSessionID() string {
	b := make([]byte, 8)
	if _, err := rand.Read(b); err != nil {
		return "mpd:unknown"
	}
	return "mpd:" + hex.EncodeToString(b)
}

func extractItem(data map[string]string) Item {
	artist := data["Artist"]
	if artist == "" {
		artist = data["AlbumArtist"]
	}
	return Item{
		Track:  data["Title"],
		Artist: artist,
		Album:  data["Album"],
		ProviderIds: map[string]string{
			"musicbrainz.track": data["MUSICBRAINZ_TRACKID"],
		},
	}
}

func postPayload(event string, item Item) {
	if dispatchURL == "" {
		slog.Error("MPD: BEETBRAINZ_URL_MPD is not set")
		return
	}

	req := Request{
		Event: event,
		User:  defaultUser,
		Item:  item,
	}

	data, err := json.Marshal(req)
	if err != nil {
		slog.Error("MPD: Failed to marshal payload", "error", err)
		return
	}

	resp, err := http.Post(dispatchURL, "application/json", strings.NewReader(string(data)))
	if err != nil {
		slog.Warn("MPD: Webhook delivery failed", "error", err)
		return
	}
	defer resp.Body.Close()

	if event == "playback.start" {
		slog.Info(fmt.Sprintf("✔️ MPD: 📻 `%s`", item.Track))
	} else if event == "playback.stop" {
		slog.Debug(fmt.Sprintf("✔️ MPD: 🎬 `%s`", item.Track))
	}
}

func readSection(scanner *bufio.Scanner) map[string]string {
	data := make(map[string]string)
	for scanner.Scan() {
		line := scanner.Text()
		if line == "OK" {
			break
		}
		if strings.HasPrefix(line, "ACK") {
			slog.Warn("MPD: error", "line", line)
			break
		}
		parts := strings.SplitN(line, ":", 2)
		if len(parts) == 2 {
			key := strings.TrimSpace(parts[0])
			value := strings.TrimSpace(parts[1])
			data[key] = value
		}
	}
	return data
}

func pollMPD(ctx context.Context) {
	host := mpdHost
	if host == "" {
		host = "localhost"
	}
	port := mpdPort
	if port == "" {
		port = "6600"
	}
	addr := net.JoinHostPort(host, port)

	conn, err := net.Dial("tcp", addr)
	if err != nil {
		slog.Error("MPD: connection failed", "error", err)
		return
	}
	slog.Debug("MPD: Connected", "address", addr)
	defer conn.Close()

	scanner := bufio.NewScanner(conn)
	if scanner.Scan() {
		slog.Debug("MPD: Handshake", "msg", scanner.Text())
	}

	for {
		select {
		case <-ctx.Done():
			slog.Debug("MPD: POST shutdown triggered")
			return
		default:
			fmt.Fprintf(conn, "idle player\n")
			for scanner.Scan() {
				line := scanner.Text()
				if line == "OK" {
					continue
				}
				if strings.HasPrefix(line, "changed: player") {
					fmt.Fprintf(conn, "noidle\n")
					break
				}
			}

			for scanner.Scan() {
				line := scanner.Text()
				if line == "OK" {
					break
				}
			}

			time.Sleep(150 * time.Millisecond)
			fmt.Fprintf(conn, "status\n")
			status := readSection(scanner)
			state := status["state"]

			fmt.Fprintf(conn, "currentsong\n")
			song := readSection(scanner)

			switch state {
			case "pause", "stop":
				if lastSeen.Track != "" {
					postPayload("playback.stop", lastSeen)
					lastSeen = Item{}
				}
			case "play":
				item := extractItem(song)
				if item.Track == "" || item.Artist == "" {
					slog.Debug("MPD: Skipping incomplete item", "metadata", song)
					break
				}
				if lastSeen.Track != item.Track ||
					lastSeen.ProviderIds["musicbrainz.track"] != item.ProviderIds["musicbrainz.track"] {
					if lastSeen.Track != "" {
						postPayload("playback.stop", lastSeen)
					}
					item.PlaySessionId = generateSessionID()
					postPayload("playback.start", item)
					lastSeen = item
				}
			default:
				slog.Debug("MPD: Unknown state", "state", state)
			}
		}
	}
}

func Run(ctx context.Context) {
	slog.Debug("MPD: POST tower activated")
	if dispatchURL == "" {
		slog.Error("MPD: Missing BEETBRAINZ_URL_MPD")
		return
	}
	slog.Debug("MPD: Configured POST endpoint", "url", dispatchURL, "user", defaultUser)
	pollMPD(ctx)
}
