// Copyright © 2022-2025 Gerald B. Cox
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

package mprispost

import (
	"context"
	"encoding/json"
	"fmt"
	"log/slog"
	"net/http"
	"os"
	"os/user"
	"strings"
	"time"

	"github.com/godbus/dbus/v5"
)

type Item struct {
	Type         string            `json:"Type"`
	Name         string            `json:"Name"`
	Album        string            `json:"Album"`
	AlbumArtist  string            `json:"AlbumArtist"`
	Artists      []string          `json:"Artists"`
	ExternalUrls []string          `json:"ExternalUrls"`
	ProviderIds  map[string]string `json:"ProviderIds"`
}

type Request struct {
	Event string `json:"Event"`
	User  struct {
		Name string `json:"Name"`
	} `json:"User"`
	Item Item `json:"Item"`
}

var (
	allowlistEnv = os.Getenv("MPRIS_ALLOWLIST")
	webhookURL   = os.Getenv("BEETBRAINZ_URL_MPRIS")
	defaultUser  = getUser()
	allowlist    = make(map[string]bool)
)

func getUser() string {
	u, err := user.Current()
	if err != nil {
		return "unknown"
	}
	return u.Username
}

func initAllowlist() {
	for _, name := range strings.Split(allowlistEnv, ",") {
		if trimmed := strings.TrimSpace(name); trimmed != "" {
			allowlist[trimmed] = true
		}
	}
	slog.Debug("MPRIS: Allowlist initialized", "players", allowlist)
	if len(allowlist) == 0 {
		slog.Error("MPRIS: ALLOWLIST is empty — please set a comma-separated list")
	}
}

func unwrap(val interface{}) interface{} {
	switch v := val.(type) {
	case dbus.Variant:
		return unwrap(v.Value())
	case map[string]dbus.Variant:
		m := make(map[string]interface{})
		for k, vv := range v {
			m[k] = unwrap(vv)
		}
		return m
	default:
		return v
	}
}

func keys(m map[string]dbus.Variant) []string {
	ks := make([]string, 0, len(m))
	for k := range m {
		ks = append(ks, k)
	}
	return ks
}

func getPlayer(name string) string {
	parts := strings.Split(name, ".")
	return parts[len(parts)-1]
}

func extractItem(meta map[string]interface{}) Item {
	item := Item{
		Type:         "Audio",
		ExternalUrls: []string{},
		ProviderIds:  map[string]string{},
	}
	if v, ok := meta["xesam:title"].(string); ok {
		item.Name = v
	}
	if v, ok := meta["xesam:album"].(string); ok {
		item.Album = v
	}
	if v, ok := meta["xesam:albumArtist"].([]string); ok && len(v) > 0 {
		item.AlbumArtist = v[0]
	}
	if v, ok := meta["xesam:artist"].([]string); ok {
		item.Artists = v
	}
	return item
}

func postPayload(event, service string, item Item, client *http.Client) {
	req := Request{Event: event, Item: item}
	req.User.Name = defaultUser

	data, err := json.Marshal(req)
	if err != nil {
		slog.Error("MPRIS: Failed to marshal payload", "error", err)
		return
	}

	slog.Debug("MPRIS: 📤 PostPayload queued", "event", event, "track", item.Name, "payload", string(data))

	resp, err := http.Post(webhookURL, "application/json", strings.NewReader(string(data)))
	if err != nil {
		slog.Warn("MPRIS: Dispatch failed", "error", err)
		return
	}
	defer resp.Body.Close()

	if event == "playback.start" {
		slog.Info(fmt.Sprintf("✔️ MPRIS: 📻 `%s`", item.Name))
	} else if event == "playback.stop" {
		slog.Debug(fmt.Sprintf("✔️ MPRIS: 🎬 `%s`", item.Name))
	}

}

func Run(ctx context.Context) {
	if webhookURL == "" {
		slog.Error("MPRIS: Missing BEETBRAINZ_URL_MPRIS")
		return
	}

	initAllowlist()
	slog.Debug("MPRIS: POST tower activated", "user", defaultUser, "url", webhookURL)

	conn, err := dbus.ConnectSessionBus()
	if err != nil {
		slog.Error("MPRIS: Failed to connect to D-Bus", "error", err)
		return
	}
	defer conn.Close()

	match := "type='signal',interface='org.freedesktop.DBus.Properties',member='PropertiesChanged'"
	call := conn.BusObject().Call("org.freedesktop.DBus.AddMatch", 0, match)
	if call.Err != nil {
		slog.Error("MPRIS: Failed to register signal match", "error", call.Err)
		return
	}

	sigs := make(chan *dbus.Signal, 128)
	conn.Signal(sigs)

	lastSeen := make(map[string]Item)
	seenSenders := make(map[string]bool)
	client := &http.Client{Timeout: 2 * time.Second}

	for {
		select {
		case <-ctx.Done():
			slog.Debug("MPRIS: POST shutdown triggered")
			return
		case sig := <-sigs:
			slog.Debug("MPRIS: 📡 Raw signal dump", "sender", sig.Sender, "path", sig.Path, "member", sig.Name, "body", sig.Body)

			if sig == nil || len(sig.Body) < 2 {
				slog.Debug("MPRIS: ⚠️ Signal is nil or body too short", "bodyLen", len(sig.Body))
				continue
			}

			if !seenSenders[sig.Sender] {
				slog.Debug("MPRIS: New D-Bus sender detected", "sender", sig.Sender)
				seenSenders[sig.Sender] = true
			}

			nameCall := conn.BusObject().Call("org.freedesktop.DBus.ListNames", 0)
			if nameCall.Err != nil {
				slog.Debug("MPRIS: ⚠️ Failed to list D-Bus names", "error", nameCall.Err)
				continue
			}

			var resolvedName string
			for _, v := range nameCall.Body {
				names, ok := v.([]string)
				if !ok {
					continue
				}
				for _, name := range names {
					if strings.HasPrefix(name, "org.mpris.MediaPlayer2.") {
						ownerCall := conn.BusObject().Call("org.freedesktop.DBus.GetNameOwner", 0, name)
						if ownerCall.Err != nil {
							continue
						}
						if ownerID, ok := ownerCall.Body[0].(string); ok && ownerID == sig.Sender {
							resolvedName = name
							break
						}
					}
				}
			}

			if resolvedName == "" {
				slog.Debug("MPRIS: 🚫 Unable to resolve player name for sender", "sender", sig.Sender)
				continue
			}

			player := getPlayer(resolvedName)
			if !allowlist[player] {
				slog.Debug("MPRIS: 🚫 Player not in allowlist", "player", player)
				continue
			}

			iface, ok := sig.Body[0].(string)
			if !ok || iface != "org.mpris.MediaPlayer2.Player" {
				slog.Debug("MPRIS: 🚫 Interface mismatch or invalid", "got", sig.Body[0])
				continue
			}

			propsRaw := sig.Body[1]
			props, ok := propsRaw.(map[string]dbus.Variant)
			if !ok {
				slog.Debug("MPRIS: ❌ Could not assert props as map[string]dbus.Variant", "props", propsRaw)
				continue
			}

			if statVar, ok := props["PlaybackStatus"]; ok {
				if status, ok := statVar.Value().(string); ok && status == "Stopped" {
					if last, ok := lastSeen[sig.Sender]; ok {
						postPayload("playback.stop", sig.Sender, last, client)
						delete(lastSeen, sig.Sender)
					}
					continue
				}
			}

			if metaVar, ok := props["Metadata"]; ok {
				metaRaw := unwrap(metaVar)
				meta, ok := metaRaw.(map[string]interface{})
				if !ok {
					slog.Debug("MPRIS: ❌ Metadata not map[string]interface{}", "raw", metaRaw)
					continue
				}

				item := extractItem(meta)
				if item.Name == "" {
					slog.Debug("MPRIS: 🚫 Item has no title — skipping dispatch", "meta", meta)
					continue
				}

				last, exists := lastSeen[sig.Sender]
				if !exists || last.Name != item.Name {
					if exists {
						postPayload("playback.stop", sig.Sender, last, client)
					}
					postPayload("playback.start", sig.Sender, item, client)
					lastSeen[sig.Sender] = item
				}
			}
		}
	}
}
