package main

import (
	"context"
	"encoding/json"
	"flag"
	"fmt"
	"log/slog"
	"net/http"
	"os"
	"os/signal"
	"strings"
	"syscall"

	"codeberg.org/gbcox/beetpost/mpdpost"
	"codeberg.org/gbcox/beetpost/mprispost"
)

const LevelStatus = slog.Level(2) // Between INFO (0) and WARN (4)

var (
	AppVersionFull  = "dev"
	CommitFull      = "none"
	ShortCommitFull = "none"
)

type Item struct {
	Track         string            `json:"Track"`
	Artist        string            `json:"Artist"`
	Album         string            `json:"Album"`
	ProviderIds   map[string]string `json:"ProviderIds"`
	PlaySessionId string            `json:"PlaySessionId"`
}

type Request struct {
	Event string `json:"Event"`
	User  string `json:"User"`
	Item  Item   `json:"Item"`
}

type PlainHandler struct {
	Level slog.Level
}

func (h *PlainHandler) Enabled(_ context.Context, level slog.Level) bool {
	return level >= h.Level
}

func (h *PlainHandler) Handle(_ context.Context, r slog.Record) error {
	var prefix string
	switch r.Level {
	case slog.LevelDebug:
		prefix = "🪲"
	case slog.LevelWarn:
		prefix = "⚠️"
	case slog.LevelError:
		prefix = "❌"
	}

	fmt.Println(prefix + " " + r.Message)
	return nil
}

func (h *PlainHandler) WithAttrs(_ []slog.Attr) slog.Handler {
	return h
}

func (h *PlainHandler) WithGroup(_ string) slog.Handler {
	return h
}

func initLogging() {
	var level slog.Level
	switch strings.ToLower(os.Getenv("BEETPOST_LOG_LEVEL")) {
	case "debug":
		level = slog.LevelDebug
	case "warn":
		level = slog.LevelWarn
	case "error":
		level = slog.LevelError
	default:
		level = slog.LevelInfo
	}

	// Set the custom plain handler as the default
	slog.SetDefault(slog.New(&PlainHandler{Level: level}))

	slog.Info(fmt.Sprintf("🎚 Logging level set — %s", level.String()))
	slog.Info(fmt.Sprintf("📮 Starting Beetpost — version: %s", AppVersionFull))

}

func SendPayload(source, event, user string, item Item) {
	url := os.Getenv("BEETBRAINZ_URL_" + strings.ToUpper(source))
	if url == "" {
		slog.Warn("⛔ No endpoint configured for source", "source", source)
		return
	}

	req := Request{
		Event: event,
		User:  user,
		Item:  item,
	}

	data, err := json.Marshal(req)
	if err != nil {
		slog.Error("💥 Payload marshal failed", "error", err)
		return
	}

	resp, err := http.Post(url, "application/json", strings.NewReader(string(data)))
	if err != nil {
		slog.Warn("📭 Dispatch failed", "source", source, "error", err)
		return
	}
	defer resp.Body.Close()

	slog.Info("📤 Payload posted", "source", source, "track", item.Track, "status", resp.StatusCode)
}

func main() {
	versionFlag := flag.Bool("version", false, "Print version information and exit")
	flag.Parse() // Parse the command-line arguments

	if *versionFlag {
		fmt.Printf("Beetpost Version: %s\n", AppVersionFull)
		fmt.Printf("Git Commit: %s (short: %s)\n", CommitFull, ShortCommitFull)
		os.Exit(0)
	}

	initLogging()

	ctx, cancel := context.WithCancel(context.Background())
	defer cancel()

	mpdEnabled := strings.ToLower(os.Getenv("MPDPOST_ENABLE")) == "true"
	mprisEnabled := strings.ToLower(os.Getenv("MPRISPOST_ENABLE")) == "true"

	if !mpdEnabled && !mprisEnabled {
		slog.Warn("🔇 No towers enabled — exiting")
		os.Exit(0)
	}

	slog.Info("🎙️ Beetpost activated")

	if mpdEnabled {
		slog.Info("🛰  MPDPOST tower enabled")
		go mpdpost.Run(ctx)
	}
	if mprisEnabled {
		slog.Info("📡 MPRISPOST tower enabled")
		go mprispost.Run(ctx)
	}

	slog.Info("📮🛰 📡 Beetpost online and ready to dispatch playback events")

	quit := make(chan os.Signal, 1)
	signal.Notify(quit, syscall.SIGINT, syscall.SIGTERM)
	sig := <-quit
	slog.Info("🛑 Shutdown signal received", "signal", sig.String())

	cancel()
	slog.Info("📮 Dispatcher shut down cleanly")
}
